# CMISS (cm) command file for full ventricular mechanics.  
#
# Prolate spheroidal coordinates are used to define the mesh consists
# using 120 elements with tricubicHermite interpolation for lambda and
# trilinear interpolation for mu, theta and the hydrostatic pressure.
# Fibre and sheet angles are interpolated with bilinear-cubicHermite
# and bicubicHermite-linear bases, resp. (imbrication angles are all
# zero).  Equations of finite deformation elasticity govern the
# boundary value problem and the material response is described by the
# pole-zero constitutive law, with prescribed residual strains from
# Omens et al (??).  Active fibre tension is modelled using the
# Hunter-McCulloch-ter Keurs equations for active myocyte contraction
# (n.b. zero for passive inflation).  LV and RV cavities (endocardial
# surfaces) are inflated to 1.0 and 0.2 kPa, resp., while lambda and
# its in-wall-plane derivs are fixed for all epicardial nodes.
#
# Created: Martyn Nash October 2000
#
 

set echo;
#==========================================================================================================
print "\033[0;30;42m ============================================================ \033[0m\n"; 
print "\033[0;30;42m                       Execution Flow                         \033[0m\n"; 
print "\033[0;30;42m ============================================================ \033[0m\n"; 
#==========================================================================================================
set echo;
#
set num_threads 12;
#
fem define parameters;r;full_big;
#
$OUTPUT_DIR = "output/";
if( ! -d ${OUTPUT_DIR})
{
		mkdir ${OUTPUT_DIR};
}
#
$TRUE = 1;
$FALSE = 0;
#
# Solve for or read in deformations
#
$MAX_PRESS = 10;
$SOLVE_INFLATION = $TRUE;
$SOLVE_ISOVOLUMIC_CONTRACTION = $TRUE;
$SOLVE_EJECTION = $TRUE;
$SOLVE_ISOVOLUMIC_RELAXATION = $FALSE;
#
if ( $SOLVE_ISOVOLUMIC_CONTRACTION || $SOLVE_EJECTION || $SOLVE_ISOVOLUMIC_RELAXATION )
{
		$COUPLED_PROBLEM = $TRUE;
} else {
		$COUPLED_PROBLEM = $FALSE;
}


set echo;
#==========================================================================================================
print "\033[0;30;42m ============================================================ \033[0m\n"; 
print "\033[0;30;42m                    Regions & Classes                         \033[0m\n"; 
print "\033[0;30;42m ============================================================ \033[0m\n"; 
#==========================================================================================================
set echo;
#
$WALL = 1;
$LV_CAVITY = 2;
$RV_CAVITY = 3;


set echo;
#==========================================================================================================
print "\033[0;30;42m ============================================================ \033[0m\n"; 
print "\033[0;30;42m                          Geometry                            \033[0m\n"; 
print "\033[0;30;42m ============================================================ \033[0m\n"; 
#==========================================================================================================
set echo;
#
fem define coord;r;fullheart_refmu region $WALL;
fem define node;r region $WALL;
fem define base;r;fullheart;
fem define element;r;fullheart_refmu region $WALL;
fem define fibre;r;fullheart_refmu region $WALL;
fem define element;r;fullheart_refmu fibre region $WALL;

#
#
# Define element groups
#
fem group elements 63..68 as base_lvfree_epi region $WALL;
fem group elements 93..98 as base_lvfree_endo region $WALL;
fem group elements 91,92,99,100 as base_septal region $WALL;
fem group elements 61,62,69,70 as base_rv region $WALL;
fem group elements base_lvfree_epi,base_lvfree_endo,base_septal,base_rv as base region $WALL;
#
fem group elements 3..8 as equatora_lvfree_epi region $WALL;
fem group elements 33..38 as equatora_lvfree_endo region $WALL;
fem group elements 31,32,39,40 as equatora_septal region $WALL;
fem group elements 1,2,9,10 as equatora_rv region $WALL;
fem group elements equatora_lvfree_epi,equatora_lvfree_endo,equatora_septal,equatora_rv as equatora region $WALL;
#
fem group elements 73..78 as equatorb_lvfree_epi region $WALL;
fem group elements 103..108 as equatorb_lvfree_endo region $WALL;
fem group elements 101,102,109,110 as equatorb_septal region $WALL;
fem group elements 71,72,79,80 as equatorb_rv region $WALL;
fem group elements equatorb_lvfree_epi,equatorb_lvfree_endo as equatorb_lvfree region $WALL;
fem group elements equatorb_lvfree,equatorb_septal,equatorb_rv as equatorb region $WALL;
#
fem group elements 13..18 as equatorc_lvfree_epi region $WALL;
fem group elements 43..48 as equatorc_lvfree_endo region $WALL;
fem group elements 41,42,49,50 as equatorc_septal region $WALL;
fem group elements 11,12,19,20 as equatorc_rv region $WALL;
fem group elements equatorc_lvfree_epi,equatorc_lvfree_endo as equatorc_lvfree region $WALL;
fem group elements equatorc_lvfree,equatorc_septal,equatorc_rv as equatorc region $WALL;
#
fem group elements equatora_lvfree_epi,equatorb_lvfree_epi,equatorc_lvfree_epi as equator_lvfree_epi region $WALL;
fem group elements equatora_lvfree_endo,equatorb_lvfree_endo,equatorc_lvfree_endo as equator_lvfree_endo region $WALL;
fem group elements equatora_septal,equatorb_septal,equatorc_septal as equator_septal region $WALL;
fem group elements equatora_rv,equatorb_rv,equatorc_rv as equator_rv region $WALL;
fem group elements equator_lvfree_epi,equator_lvfree_endo,equator_septal,equator_rv as equator region $WALL;
#
fem group elements 81..90 as apexa_epi region $WALL;
fem group elements 111..120 as apexa_endo region $WALL;
fem group elements apexa_epi,apexa_endo as apexa region $WALL;
#
fem group elements 21..30 as apexb_epi region $WALL;
fem group elements 51..60 as apexb_endo region $WALL;
fem group elements apexb_epi,apexb_endo as apexb region $WALL;
#
fem group elements apexa_epi,apexb_epi as apex_epi_elems region $WALL;
fem group elements apexa_endo,apexb_endo as apex_endo_elems region $WALL;
fem group elements apex_endo_elems,apex_epi_elems as apex_elems region $WALL;
#
fem group elements base_lvfree_endo,equator_lvfree_endo as lvfree_endo_elems region $WALL;
fem group elements base_lvfree_epi,equator_lvfree_epi as lvfree_epi_elems region $WALL;
fem group elements base_rv,equator_rv as rv_elems region $WALL;
fem group elements base_septal,equator_septal as septal_elems region $WALL;


#
#
# Define node groups
#
fem group nodes xi2=1 xi3=1 external elements base_lvfree_epi,base_rv as base_epi_nodes region $WALL;
fem group nodes xi2=1 xi3=0 external elements base_lvfree_endo,base_septal as base_lv_endo_nodes region $WALL;
fem group nodes xi2=1 xi3=0 external elements base_rv as base_rvfree_endo_nodes region $WALL;
fem group nodes xi2=1 xi3=1 external elements base_septal as base_rvsept_endo_nodes region $WALL;
fem group nodes base_lv_endo_nodes,base_rvfree_endo_nodes,base_rvsept_endo_nodes as base_endo_nodes region $WALL;
fem group nodes xi2=1 external elements base as base_nodes region $WALL;
#
# LV endo node groups
fem group nodes xi2=1 xi3=0 external elements equatora_lvfree_endo,equatora_septal as eqa_lv_endo_nodes region $WALL;
fem group nodes xi2=1 xi3=0 external elements equatorb_lvfree_endo,equatorb_septal as eqb_lv_endo_nodes region $WALL;
fem group nodes xi2=1 xi3=0 external elements equatorc_lvfree_endo,equatorc_septal as eqc_lv_endo_nodes region $WALL;
fem group nodes xi2=1 xi3=0 external elements apexa_endo as apexa_lv_endo_nodes region $WALL;
fem group nodes xi2=1 xi3=0 external elements apexb_endo as apexb_lv_endo_nodes region $WALL;
fem group nodes eqa_lv_endo_nodes,eqb_lv_endo_nodes,eqc_lv_endo_nodes,apexa_lv_endo_nodes as equator_lv_endo_nodes region $WALL;
#
# Epi node groups
fem group nodes xi2=1 xi3=1 external elements equatora_lvfree_epi,equatora_rv as eqa_epi_nodes region $WALL;
fem group nodes xi2=1 xi3=1 external elements equatorb_lvfree_epi,equatorb_rv as eqb_epi_nodes region $WALL;
fem group nodes xi2=1 xi3=1 external elements equatorc_lvfree_epi,equatorc_rv as eqc_epi_nodes region $WALL;
fem group nodes xi2=1 xi3=1 external elements apexa_epi as apexa_epi_nodes region $WALL;
fem group nodes xi2=1 xi3=1 external elements apexb_epi as apexb_epi_nodes region $WALL;
fem group nodes eqa_epi_nodes,eqb_epi_nodes,eqc_epi_nodes,apexa_epi_nodes as equator_epi_nodes region $WALL;
#
fem group nodes xi3=0 external elements equator_rv as equator_rvfree_endo_nodes region $WALL;
fem group nodes xi3=1 external elements equator_septal as equator_rvsept_endo_nodes region $WALL;
fem group nodes equator_rvfree_endo_nodes,equator_rvsept_endo_nodes as equator_rv_endo_nodes region $WALL;
fem group nodes xi3=1 elements equator_lvfree_endo,apex_endo_elems as equator_lv_mid_nodes region $WALL;
fem group nodes xi3=1 external elements lvfree_epi_elems,rv_elems,apex_epi_elems as epi_nodes region $WALL;
#
fem group nodes xi2=0 elements apexb as apex_nodes region $WALL;
fem group nodes xi2=0 xi3=0 external elements apexa_endo as near_apex_endo_nodes region $WALL;
fem group nodes xi2=0 xi3=1 external elements apexa_epi as near_apex_epi_nodes region $WALL;
fem group nodes equator_lv_mid_nodes,equator_rv_endo_nodes as equator_mid_nodes_temp region $WALL;
fem group nodes equator_mid_nodes_temp,near_apex_endo_nodes,near_apex_epi_nodes as equator_mid_nodes region $WALL;


#
set echo;
#==========================================================================================================
print "\033[0;30;42m ============================================================ \033[0m\n"; 
print "\033[0;30;42m           Dependent variable/material information            \033[0m\n"; 
print "\033[0;30;42m ============================================================ \033[0m\n"; 
#==========================================================================================================
set echo;
#
fem define equation;r;finelas_tch_incomp region $WALL lock;
fem define material;r;orth_incomp_active_residstrain region $WALL;
fem define active;r;active0_00 region $WALL;
fem define initial;r;fullheart_inflate_pericard region $WALL;
fem define solve;r;newton region $WALL;


set echo; 
#==========================================================================================================
print "\033[0;30;42m ============================================================ \033[0m\n"; 
print "\033[0;30;42m                           Inflation                          \033[0m\n"; 
print "\033[0;30;42m ============================================================ \033[0m\n"; 
#==========================================================================================================
if ( $SOLVE_INFLATION )
{ 
	 print "\033[0;30;43m Increase cavity pressures incrementally to simulate diastole \033[0m\n"; 
   #
   set output;${OUTPUT_DIR}fullheart_inflate on;
   $NAME1 = "fullheart_def_00";
   $NAME2 = "_active_0000";
   $ITERS = 50;
   $ERROR_TOLERANCE = 0.001;
   #
   for $PRESS ( 0..$MAX_PRESS ) 
   {
			 print "\033[0;30;43m             Pressure = $PRESS              \033[0m\n"; 
			 $NUMZEROS=2-length($PRESS);
			 $PRESSNUM="";
			 for $Z ( 1..$NUMZEROS ) 
			 {
					 $PRESSNUM=$PRESSNUM."0";
			 }
			 $PRESSNUM=$PRESSNUM.$PRESS;
			 if( $PRESS == 0 )
			 {
					 $INCREM = 0.0;
			 } else
			 {
					 $INCREM = 0.1;
			 }
			 fem solve increment $INCREM iterate $ITERS error $ERROR_TOLERANCE;
			 $FILENAME = ${NAME1}.${PRESSNUM}.${NAME2};
       fem define initial;w;${OUTPUT_DIR}.${FILENAME} region $WALL;
       fem export nodes;${OUTPUT_DIR}.${FILENAME} as heart;
       fem export nodes;${OUTPUT_DIR}${FILENAME}_def field as heart;
       fem export elements;${OUTPUT_DIR}.${FILENAME} as wall;
       fem export elements;${OUTPUT_DIR}${FILENAME}_def field as wall;
    }   
    set output off;

} else {
	 print "\033[0;30;43m                Reading in end-diastolic state                \033[0m\n"; 
   #
	 $FILENAME = "fullheart_def_00".${MAX_PRESS}."_active_0000";
	 fem define initial;r;${OUTPUT_DIR}.${FILENAME} region $WALL;
   fem define solve;r;newton region $WALL;
   fem export nodes;${OUTPUT_DIR}.${FILENAME} as heart;
   fem export nodes;${OUTPUT_DIR}${FILENAME}_def field as heart;
   fem export elements;${OUTPUT_DIR}.${FILENAME} as wall;
   fem export elements;${OUTPUT_DIR}${FILENAME}_def field as wall;
}
set echo;



if ( $COUPLED_PROBLEM ) {           # Set up coupled regions
    set echo;
    #==========================================================================================================
		print "\033[0;30;42m ============================================================ \033[0m\n"; 
		print "\033[0;30;42m                    Set up Coupled Regions                    \033[0m\n"; 
		print "\033[0;30;42m ============================================================ \033[0m\n"; 
    #==========================================================================================================
    #
    # Save scale factors for the ventricular wall elements
    #
		fem define base;r;fullheart_readse;              #redefine bases to read in scale factors 
		fem def line;w;fullheart_refmu region $WALL;     #write out scale factors for wall elements
		fem define base;r;fullheart;                     #redefine bases to calculate scale factors
    #
    # Define geometry and dependent variable information for LV and RV cavities
    #
		fem define region;r;coupled;
		fem define coord;r;lvcavity_refmu region $LV_CAVITY;
		fem define node;r region $LV_CAVITY;
		fem define coord;r;rvcavity_refmu region $RV_CAVITY;
		fem define node;r region $RV_CAVITY;
		fem define element;r;lvcavity_refmu region $LV_CAVITY;
		fem define element;r;rvcavity_refmu region $RV_CAVITY;
		fem group elem 121..180 as lv_cavity_elems region $LV_CAVITY;
		fem group elem 151..160 as lv_cavity_base_elems region $LV_CAVITY;
		fem group nodes xi2=1 xi3=1 elem lv_cavity_base_elems as lv_cavity_base_nodes region $LV_CAVITY;
		fem group elem 181..198 as rv_cavity_elems region $RV_CAVITY;
		fem define equation;r;coupled region $WALL,$LV_CAVITY,$RV_CAVITY lock;
		fem define material;r;cavity region $LV_CAVITY;
		fem define material;r;cavity region $RV_CAVITY;
		fem define initial;r;lvcavity_refmu region $LV_CAVITY;
		fem define initial;r;rvcavity_refmu region $RV_CAVITY;
    #
    # Define cavity/wall coupling and solution information
    #
		fem define coupling;r;coupled;
		fem define solve;r;coupled coupled region $WALL,$LV_CAVITY,$RV_CAVITY;
		fem update solution coupled source_region $WALL;
		$NHlambda = 1;
		fem update solution cavity_reference average 196 in $NHlambda node lv_cavity_base_nodes region $LV_CAVITY;
		fem update solution cavity_reference average 203 in $NHlambda node 14,18 region $RV_CAVITY;
    #
    # Read back in scale factors for the ventricular wall elements
    #
		fem define base;r;fullheart_readse;           #redefine bases to read in scale factors 
		fem def line;r;fullheart_refmu region $WALL;  #read in scale factors for wall elements
		fem define base;r;fullheart;                  #redefine bases to calculate scale factors
		fem export nodes;${OUTPUT_DIR}heart as heart region all;          
		fem export elements;${OUTPUT_DIR}lv_cavity as lv_cavity region $LV_CAVITY;
		fem export elements;${OUTPUT_DIR}rv_cavity as rv_cavity region $RV_CAVITY;
		fem export elements;${OUTPUT_DIR}wall as wall region $WALL;
		fem export nodes;${OUTPUT_DIR}heart_def field as heart region all;          
		fem export elements;${OUTPUT_DIR}lv_cavity_def field as lv_cavity region $LV_CAVITY;
		fem export elements;${OUTPUT_DIR}rv_cavity_def field as rv_cavity region $RV_CAVITY;
		fem export elements;${OUTPUT_DIR}wall_def field as wall region $WALL;
    set echo;
}



if ( $SOLVE_ISOVOLUMIC_CONTRACTION ) {
    set echo;
    #==========================================================================================================
    print "\033[0;30;42m ============================================================ \033[0m\n"; 
    print "\033[0;30;42m             Solving Isovolumic Contraction                   \033[0m\n"; 
    print "\033[0;30;42m ============================================================ \033[0m\n"; 
    #==========================================================================================================
    set output;${OUTPUT_DIR}steps_contract_infl0_${MAX_PRESS} on;
    $NAME1 = "active0_";
    $NAME2 = "fullheart_def_00".${MAX_PRESS}."_active_00";
    $INCREM = 0.0;
    $ITER = 20;
    $ERROR_TOLERANCE = 0.001;
    #
    # Isovolumically contract the ventricles by incrementally
    # increasing intracellular calcium concentration and solving.
    if ( $MAX_PRESS == 10 ) {
				$MAX_CA = 24;
    } elsif ( $MAX_PRESS == 30 ) {
				$MAX_CA = 30;
		} else {
				fem quit # required activation level unknown
		}
		for $CA ( 1..$MAX_CA )
		{
				$NUMZEROS=2-length($CA);
				$CANUM="";
				for $Z ( 1..$NUMZEROS ) 
				{
						$CANUM=$CANUM."0";
				}
				$CANUM=$CANUM.$CA
				print "\033[0;30;43m                    Calcium Level = ${CA}                         \033[0m\n"; 
		    $FILENAME=${NAME2}.${CANUM}
				fem define active;r;${NAME1}.${CANUM} region $WALL;
		    fem solve increment $INCREM iter $ITER coupled error $ERROR_TOLERANCE;
		    fem define initial;w;${OUTPUT_DIR}.${FILENAME} region $WALL,$LV_CAVITY,$RV_CAVITY;
        fem export nodes;${OUTPUT_DIR}.${FILENAME} as heart region all;
        fem export nodes;${OUTPUT_DIR}${FILENAME}_def field as heart region all;
    }
    set out off
    set echo;
}



if ( $SOLVE_EJECTION ) {
		set echo;
    #==========================================================================================================
		print "\033[0;30;42m ============================================================ \033[0m\n"; 
		print "\033[0;30;42m                   Solving Ejection                           \033[0m\n"; 
		print "\033[0;30;42m ============================================================ \033[0m\n"; 
    #==========================================================================================================
		if ( !$SOLVE_ISOVOLUMIC_CONTRACTION ) {
 				print "\033[0;30;43m    Reading in state at the end of isovolumic contraction     \033[0m\n"; 
				$CA = 24;
				fem define active;r;"active0_".${CA} region $WALL;
				fem define initial;r;${OUTPUT_DIR}."fullheart_def_00".${MAX_PRESS}."_active_00".${CA} nofix region $WALL,$LV_CAVITY,$RV_CAVITY;
		}
    $CA = 24;
    set out;${OUTPUT_DIR}."steps_eject_infl0_".${MAX_PRESS}."_act0_".${CA} on;
		$NAME1 = "cavity_";
    $NAME2 = "fullheart_def_00".${MAX_PRESS}."_active_00".${CA}."_cavk_";
		$INCREM = 0.0;
		$ITERS = 20;
		$ERROR_TOLERANCE = 0.001;
    #
    # Simulate ventricular ejection by incrementally decreasing the 
    # cavity impedances.
		$CAVITY_K =640;
		for $I ( 1..26 ) 
		{
				$CAVITY_K2 = int($CAVITY_K*1000);
				$NUMZEROS=8-length($CAVITY_K2)
						$CAVITY_K_NUMBER="";
				for $Z ( 1..$NUMZEROS ) 
				{
						$CAVITY_K_NUMBER=$CAVITY_K_NUMBER."0";
				}
				$CAVITY_K_NUMBER=$CAVITY_K_NUMBER.$CAVITY_K2;
		    $FILENAME=${NAME2}.${CAVITY_K_NUMBER};
				print "\033[0;30;43m          Cavity Stiffness Number   ${CAVITY_K_NUMBER}                 \033[0m\n"; 
				fem define material;r;${NAME1}.${CAVITY_K_NUMBER} region $LV_CAVITY;
        fem define material;r;${NAME1}.${CAVITY_K_NUMBER} region $RV_CAVITY;
        fem solve increment $INCREM iterate $ITERS coupled error $ERROR_TOLERANCE;
        fem define initial;w;${OUTPUT_DIR}.${NAME2}.${CAVITY_K_NUMBER} region $WALL,$LV_CAVITY,$RV_CAVITY;
        fem export nodes;${OUTPUT_DIR}.${FILENAME} as heart region all;
        fem export nodes;${OUTPUT_DIR}${FILENAME}_def field as heart region all;
        if( $CAVITY_K > 10 )
        {
            $CAVITY_K = $CAVITY_K / 2.0;
        } elsif( $CAVITY_K > 4 )
        {
            $CAVITY_K = $CAVITY_K - 1.0;
        } else
        {
            $CAVITY_K = $CAVITY_K - 0.3;
        }
    }
    set out off;
    set echo;
}



if ( $SOLVE_ISOVOLUMIC_RELAXATION ) {
set echo;
#==========================================================================================================
print "\033[0;30;42m ============================================================ \033[0m\n"; 
print "\033[0;30;42m              Solving Isovolumic Relaxation                   \033[0m\n"; 
print "\033[0;30;42m ============================================================ \033[0m\n"; 
#==========================================================================================================
set echo;
}















