#Example_991  Preliminary compressible mechanics for lung example.

# Simple example of the lung compressible mechanics solution. Here we 
# use Voronoi cells to mesh the volume inside the element, but in future
# work we will be using a more simple description based on a regular grid
# of points (faster, less memory overhead, relates directly to generation
# of the airway mesh). The current example does not have an embedded
# airway mesh and flow solution. I have indicated the places that this 
# would be done though.

$tissue_region=1;
$embed_region=2;

$pressure_field=1;
$init_vol_field=2;
$crnt_vol_field=3;

fem def para;r;big;
fem def coor 3,1;
fem def base;r;comb2;  # 2 linear for mechanics, 2 simplex for voronoi
fem def regions;r;comb;  # reg. 1 is mechanics, reg. 2 is voronoi

#DEFINE HOST GEOMETRY
#simple tri-linear cube
fem def node;r;cube reg $tissue_region;
fem def elem;r;cube reg $tissue_region;

fem group nodes at xi1=0 as xi1_0 reg $tissue_region;
fem group nodes at xi1=1 as xi1_1 reg $tissue_region;
fem group nodes at xi2=0 as xi2_0 reg $tissue_region;
fem group nodes at xi2=1 as xi2_1 reg $tissue_region;
fem group nodes at xi3=0 as xi3_0 reg $tissue_region;
fem group nodes at xi3=1 as xi3_1 reg $tissue_region;

#DEFINE DELAUNAY-VORONOI GEOMETRY
#  calculate the location of boundary and internal boundary nodes to define surface
fem def node;c delaunay reg $tissue_region target 2 division 2 radius 0.05 internal;
#  group nodes internal to the simple (mechanics) element as 
#  points to which the pressure field will be fit. for the coupled airway model 
#  the points would be the terminal nodes.
fem group nodes internal_boundary,internal as pressure_nodes reg $embed_region;
#  calculate a Delaunay triangulation of the points
fem def mesh;c delaunay basis 3 B boundary IB internal_boundary IN_nodes internal reg $embed_region;
fem update delaunay reg $embed_region;
#  calculate the Voronoi cells (dual of Delaunay triangulation)
fem def mesh;c voronoi B boundary IB internal_boundary IN_nodes internal reg $embed_region;

#SET UP FOR MECHANICS SOLUTION
fem de fibre;d reg $tissue_region;
fem de elem;d fibre reg $tissue_region;

#fem def equa;r;Lee;  # use compressible
fem def equa;r;Lee_Comp_fluid;  # use compressible + fluid
fem def mate;r;Lee reg $tissue_region;  # use Lee's strain energy function

fem define field;d reg $tissue_region field_variable 1; 
#  set the initial pressures all = 1. the lung model would all = 0 during zero flow.
#fem update field $pressure_field reg $tissue_region at nodes add constant 1.0;
fem def elem;r;cube field reg $tissue_region;

fem define field;d reg $embed_region field_variable 3; #pressure, previous volume,current volume fields
#  set the initial pressures all = 1. the lung model would all = 0 during zero flow.
fem update field $pressure_field reg $embed_region at nodes pressure_nodes add constant 1.0;

fem de elem;d field reg $embed_region;

fem def init;r;dispcube reg $tissue_region;
fem def fit;r;linear field node_group reg $tissue_region;
fem def solv;r;linear reg $tissue_region;
fem def xi;c nodes of $embed_region in $tissue_region; #XI of Delaunay nodes in host geometry
fem de xi;c vertices of $embed_region in $tissue_region; #XI of Voronoi vertices in host geometry
fem update voronoi reg $embed_region field $crnt_vol_field; #calculates Voronoi cell volumes, puts into fields
fem update field $init_vol_field reg $embed_region at nodes pressure_nodes subst field $crnt_vol_field;

fem list voronoi;undeformed reg $embed_region;
fem solve step 1 reg $tissue_region;
fem list elem deform total reg $tissue_region basis 1;
for $COUNT (1..5)
{
    #CALCULATE GLOBAL COORDINATES OF EMBEDDED NODES AND VERTICES
    fem update mesh position of $embed_region in $tissue_region deform contain; #updates Delaunay node positions
    fem update vertices of $embed_region in $tissue_region deform; #updates Voronoi vertex positions
    
    #CALCULATE PRESSURES AT EMBEDDED NODES (P1V1=P2V2)
    #only update the initial volume field after update geometry from solution
    #fem update field $init_vol_field reg $embed_region at nodes nodes pressure_nodes subst field $crnt_vol_field;
    fem update voronoi reg $embed_region field $crnt_vol_field; #recalculates Voronoi cell volumes
    #THIS IS WHERE THE AIRWAY FLOW SOLUTION WILL GO. Solve with flow BCs for terminal pressures.
    #Currently just using simple pressure-volume relationship to update local pressures.
    
    fem update field ${pressure_field} reg $embed_region at nodes pressure_nodes multiply field ${init_vol_field}; #=P1*V1
    fem update field ${pressure_field} reg $embed_region at nodes pressure_nodes divide field ${crnt_vol_field}; #=P1*V1/V2
    
    #FIT PRESSURES AT EMBEDDED NODES TO NODES IN TISSUE REGION
    fem fit reg $tissue_region; #fit pressures at pressure_nodes (region 2) to nodes in region 1
    fem de field;w;iteratefile reg $tissue_region; #write out the pressure field
    fem de init;r;dispcube reg $tissue_region; #re-read initial conditions
    fem de solve;r;linear reg $tissue_region; #re-read solve
    fem de field;r;iteratefile reg $tissue_region; #read in previous fit field
    fem solve step 1 reg $tissue_region; #solve with updated pressures
    fem list elem deform total reg $tissue_region;
    fem read com;update_fit_solve;
}

fem export node;pressure as pressure nodes pressure_nodes reg 2;
fem quit;

